/*------------------------------------------------------------------------------
 bin/djv_ls/djv_ls.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_base/math.h>
#include <djv_base/user.h>
#include <djv_base/sys.h>
#include <djv_base/file.h>
#include <djv_base/dir.h>
#include <djv_base/app.h>

using namespace djv;

class App : public app::App {
public:

  App(int, char **);
  ~App();

private:

  void file(const File &, bool path = false);
  bool dir(const File &, bool label);
  
  bool cmdln(int, char **) throw (Error);
  Str help() const;

  List<Str> _input;
  bool _seq, _recurse, _hidden;
  List<Str> _filter;
  int _columns;
  file::SORT _sort;
  bool _reverse, _dirs;
};

App::App(int argc, char ** argv) :
  app::App("djv_ls", argc, argv),
  _seq(true), _recurse(false), _hidden(false), _columns(sys::terminal_width()),
  _sort(file::SORT_NAME), _reverse(false), _dirs(true)
{
  DJV_DBG_OBJ("App::App");
  
  // Command line.
  
  bool r = false;
  try { r = cmdln(argc, argv); }
  catch (Error in) { error(str::Format("Command Line: %1").arg(in)); }
  if (! r) { exit(1); return; }
  
  // Setup.

  List<File> list;
  for (uint i = 0; i < _input.size(); ++i) {
    File file(_input[i], false);
    if (! file.stat()) {
      this->error(str::Format("Open: %1").arg(file.name()));
      r = false;
      continue;
    }
    if (_seq && file.seq_valid()) file.type = file::SEQ;
    list += file;
  }

  if (_seq) file::seq_compress(&list);
  if (! _hidden) file::filter(&list, file::FILTER_HIDDEN, _filter);
  file::sort(&list, _sort, _reverse);
  if (_dirs) list = file::sort_dirs(list);

  if (! list.size() && r)
    list += File(".");
  
  // Work.

  for (uint i = 0; i < list.size(); ++i)
    if (file::DIR == list[i].type)
      r &= dir(list[i], (list.size() > 1) || _recurse);
    else
      file(list[i], true);
  exit(! r);
}

App::~App()
{
  DJV_DBG_OBJ("App::~App");
}

void App::file(const File & in, bool path)
{
  DJV_DBG_OBJ("App::file");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("path = " << path);
  DJV_DBG_PRN("columns = " << _columns);
  
  Str name = in.get(-1, path);
  
  const Str info_str =
#if defined(WIN32)
    str::Format("%1 %2 %3 %4").
#else
    str::Format("%1 %2 %3 %4 %5").
#endif
    arg(file::size_label(in.size)).
#if ! defined(WIN32)
    arg(user::uid_to_str(in.user)).
#endif
    arg(file::type_label()[in.type], 4).
    arg(file::perm_label(in.perm), 3).
    arg(djv::time::time_to_str(in.time));

  if (_columns) {
    const uint size = math::max(_columns - int(info_str.size()) - 2, 0);
    if (name.size() > size) {
      name = Str(name, 0, size);
      size_t i = name.size(), j = 3;
      while (i-- > 0 && j-- > 0) name[i] = '.';
    }
  }
  
  print(str::Format("%1 %2").arg(name).
    arg(info_str, _columns - int(name.size()) - 2));
}

bool App::dir(const File & in, bool label)
{
  DJV_DBG_OBJ("App::dir");
  DJV_DBG_PRN("in = " << in);
  
  // Setup.

  Dir dir(in);
  if (! dir.valid()) {
    error(str::Format("Open: %1").arg(in.name()));
    return false;
  }
  List<File> list = dir.list();
  
  if (_seq) file::seq_compress(&list);
  if (! _hidden) file::filter(&list, file::FILTER_HIDDEN, _filter);
  file::sort(&list, _sort, _reverse);
  if (_dirs) list = file::sort_dirs(list);

  // Print.

  if (label) print(str::Format("%1:").arg(in));
  for (uint i = 0; i < list.size(); ++i)
    file(list[i]);
  separator();
  
  // Recurse.

  bool r = true;
  if (_recurse) {
    Dir dir(in);
    List<File> list = dir.list();
    file::filter(&list, file::FILTER_FILE |
      (! _hidden ? file::FILTER_HIDDEN : 0));
    for (uint i = 0; i < list.size(); ++i)
      r &= this->dir(list[i], label);
  }
  return r;
}

bool App::cmdln(int argc, char ** argv) throw (Error)
{
  if (! djv::app::App::cmdln(argc, argv)) return false;
  
  List<Str> list;
  for (int i = 1; i < argc; ++i) list += argv[i];
  
  while (list.size()) {
    const Str arg = list.pop_front();

    // Options.

    if (str::compare("-x_seq", arg) || str::compare("-q", arg)) {
      _seq = false;
    } else if (str::compare("-recurse", arg) || str::compare("-r", arg)) {
      _recurse = true;
    } else if (str::compare("-hidden", arg) || str::compare("-n", arg)) {
      _hidden = true;
    } else if (str::compare("-filter", arg) || str::compare("-f", arg)) {
      Str tmp;
      if (! str::serialize(&list, &tmp)) throw Error(arg);
      _filter += tmp;
    } else if (str::compare("-columns", arg) || str::compare("-c", arg)) {
      if (! str::serialize(&list, &_columns)) throw Error(arg);
    } else if (str::compare("-sort", arg) || str::compare("-s", arg)) {
      if (! str::serialize(&list, &_sort, file::sort_label())) throw Error(arg);
    } else if (str::compare("-reverse", arg) || str::compare("-rv", arg)) {
      _reverse = true;
    } else if (str::compare("-x_dirs", arg) || str::compare("-d", arg)) {
      _dirs = false;
    }
    
    // Arguments.
    
    else _input += arg;
  }
  
  return true;
}

Str App::help() const
{
  return str::Format(
"\n"
"djv_ls\n"
"------\n"
"Directory listing utility.\n"
"\n"
"Output:\n"
"\n"
"el_cerrito_bart.1k.tiff        2.23MB darby File rw Mon Jun 12 21:21:55 2006\n"
"richmond_train.2k.tiff         8.86MB darby File rw Mon Jun 12 21:21:58 2006\n"
"fishpond.1-749.png            293.17MB darby Seq rw Thu Aug 17 16:47:43 2006\n"
"\n"
"(name) (size) (user) (type) (permissions) (time)\n"
"\n"
"Type: file, seq (file sequence), dir (directory).\n"
"Permissions: r (readable), w (writable), x (executable).\n"
"\n"
"Usage\n"
"-----\n"
"djv_ls [file|directory]... [option]...\n"
"\n"
"Options\n"
"-------\n"
"-x_seq, -q - Disable file sequencing.\n"
"-recurse, -r - Descend into sub-directories.\n"
"-hidden - Show hidden files.\n"
"-filter, -f (value) - Filter with shell globbing (e.g., *, ?).\n"
"-columns, -c (value) - Output columns. Zero disables formatting.\n"
"\n"
"Sorting Options\n"
"---------------\n"
"-sort, -s (value) - Sort: %1. Default: %2.\n"
"-reverse, -rv - Reverse sorting order.\n"
"-x_dirs, -d - Disable sorting directories first.\n"
"\n"
"%3"
"\n"
"Examples\n"
"--------\n"
"> djv_ls\n"
"List the current directory.\n"
"\n"
"> djv_ls ~/movies ~/pictures\n"
"List multiple directories.\n"
"\n"
"> djv_ls -sort size -reverse\n"
"Sort by size, largest first.\n"
"\n"
"> djv_ls -filter \"*.tiff\" -filter \"*.tif\"\n"
"Show only files ending with \".tiff\" or \".tif\".\n"
  ).
    arg(str::join(file::sort_label(), ", ")).
    arg(file::sort_label()[_sort]).
    arg(djv::app::App::help());
}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = App(argc, argv).exec(); }
  catch (Error in) { error::print(in); }
  return r;
}

