/*------------------------------------------------------------------------------
 lib/djv_image/data_fnc.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/data_fnc.cxx"

#include "data.h"

namespace djv {
namespace image {
namespace data {

size_t scanline_bytes(const Info & in)
{
  return (in.size.x * pixel::bytes(in.pixel) * in.align) / in.align;
}

size_t data_bytes(const Info & in)
{
  return in.size.y * scanline_bytes(in);
}

void proxy_scale(const Data & in, Data * out, int proxy)
{
  DJV_ASSERT(out);
  DJV_ASSERT(in.pixel() == out->pixel());
  DJV_ASSERT(proxy >= 0);
  
  DJV_DBG_OBJ2(_DBG, "proxy_scale");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("out = " << *out);
  DJV_DBG_PRN("proxy = " << proxy);
  
  const int w = out->w(), h = out->h();
  const int proxy_scale = data::proxy_scale(proxy);
  const int pixel_bytes = out->pixel_bytes();
  
  for (int y = 0; y < h; ++y)
    switch (pixel_bytes) {
    
      // XXX Alignment issues?
      
      /*case 1: {
        const uint8_t * in_p = in.data(0, y * proxy_scale);
        uint8_t * out_p = out->data(0, y);
        for (int x = 0; x < w; ++x, in_p += proxy_scale, ++out_p)
          *out_p = *in_p;
      } break;
      
      case 2: {
        const uint16_t * in_p =
          reinterpret_cast<const uint16_t *>(in.data(0, y * proxy_scale));
        uint16_t * out_p =
          reinterpret_cast<uint16_t *>(out->data(0, y));
        for (int x = 0; x < w; ++x, in_p += proxy_scale, ++out_p)
          *out_p = *in_p;
      } break;
      
      case 4: {
        const uint32_t * in_p =
          reinterpret_cast<const uint32_t *>(in.data(0, y * proxy_scale));
        uint32_t * out_p =
          reinterpret_cast<uint32_t *>(out->data(0, y));
        for (int x = 0; x < w; ++x, in_p += proxy_scale, ++out_p)
          *out_p = *in_p;
      } break;
      
      case 8: {
        const uint64_t * in_p =
          reinterpret_cast<const uint64_t *>(in.data(0, y * proxy_scale));
        uint64_t * out_p =
          reinterpret_cast<uint64_t *>(out->data(0, y));
        for (int x = 0; x < w; ++x, in_p += proxy_scale, ++out_p)
          *out_p = *in_p;
      } break;*/
      
      default: {
        const uint8_t * in_p = in.data(0, y * proxy_scale);
        uint8_t * out_p = out->data(0, y);
        for (int x = 0; x < w; ++x, in_p += pixel_bytes * proxy_scale,
          out_p += pixel_bytes
        )
          switch (pixel_bytes) {
            case 16: out_p[15] = in_p[15];
            case 15: out_p[14] = in_p[14];
            case 14: out_p[13] = in_p[13];
            case 13: out_p[12] = in_p[12];
            case 12: out_p[11] = in_p[11];
            case 11: out_p[10] = in_p[10];
            case 10: out_p[9] = in_p[9];
            case 9: out_p[8] = in_p[8];
            case 8: out_p[7] = in_p[7];
            case 7: out_p[6] = in_p[6];
            case 6: out_p[5] = in_p[5];
            case 5: out_p[4] = in_p[4];
            case 4: out_p[3] = in_p[3];
            case 3: out_p[2] = in_p[2];
            case 2: out_p[1] = in_p[1];
            case 1: out_p[0] = in_p[0];
          }
      } break;
  }
}

int proxy_scale(int proxy)
{
  return (proxy > 0) ? math::pow(2, proxy) : 1;
}

Vec2i proxy_scale(const Vec2i & in, int proxy)
{
  const int scale = proxy_scale(proxy);
  return Vec2i(
    math::ceil(in.x / double(scale)),
    math::ceil(in.y / double(scale)));
}

void planar_interleave(const Data & in, Data * out, int proxy)
{
  DJV_ASSERT(out);
  DJV_ASSERT(in.pixel() == out->pixel());
  DJV_ASSERT(in.pixel() != pixel::RGB_U10);
  
  DJV_DBG_OBJ2(_DBG, "planar_interleave");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("out = " << *out);
  DJV_DBG_PRN("proxy = " << proxy);
  
  const int w = out->w(), h = out->h();
  const int proxy_scale = data::proxy_scale(proxy);
  const int channels = out->channels();
  const int pixel_bytes = out->pixel_bytes();
  const int channel_bytes = pixel::channel_bytes(out->pixel());
  
  for (int c = 0; c < channels; ++c)
    for (int y = 0; y < h; ++y)
      switch (channel_bytes) {
        default: {
          const uint8_t * in_p = in.data() + (c * in.h() + y * proxy_scale) *
            in.w() * channel_bytes;
          uint8_t * out_p = out->data(0, y) + c * channel_bytes;
          for (int x = 0; x < w; ++x, in_p += channel_bytes * proxy_scale,
            out_p += pixel_bytes
          )
            switch (channel_bytes) {
              case 4: out_p[3] = in_p[3];
              case 3: out_p[2] = in_p[2];
              case 2: out_p[1] = in_p[1];
              case 1: out_p[0] = in_p[0];
            }
        } break;
      }
}

void planar_deinterleave(const Data & in, Data * out)
{
  DJV_ASSERT(out);
  DJV_ASSERT(in.pixel() == out->pixel());
  DJV_ASSERT(in.pixel() != pixel::RGB_U10);
  
  DJV_DBG_OBJ2(_DBG, "planar_deinterleave");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("out = " << *out);
  
  const int w = out->w(), h = out->h();
  const int channels = out->channels();
  const int pixel_bytes = out->pixel_bytes();
  const int channel_bytes = pixel::channel_bytes(out->pixel());
  
  for (int c = 0; c < channels; ++c)
    for (int y = 0; y < h; ++y)
      switch (channel_bytes) {
        default: {
          const uint8_t * in_p = in.data(0, y) + c * channel_bytes;
          uint8_t * out_p = out->data() + (c * h + y) * w * channel_bytes;
          for (int x = 0; x < w; ++x, in_p += pixel_bytes, out_p += channel_bytes)
            switch (channel_bytes) {
              case 4: out_p[3] = in_p[3];
              case 3: out_p[2] = in_p[2];
              case 2: out_p[1] = in_p[1];
              case 1: out_p[0] = in_p[0];
            }
        } break;
      }
}

bool operator == (const Info & a, const Info & b)
{
  return a.size == b.size && a.proxy == b.proxy && a.pixel == b.pixel &&
    a.bgr == b.bgr && a.mirror_x == b.mirror_x && a.mirror_y == b.mirror_y &&
    a.align == b.align && a.endian == b.endian;
}

bool operator != (const Info & a, const Info & b) { return ! (a == b); }

Dbg & operator << (Dbg & dbg, const Info & in)
{
  return dbg <<
    in.size << "/" << in.proxy << " " << in.pixel <<
    " bgr=" << in.bgr <<
    " mirror=" << in.mirror_x << "|" << in.mirror_y <<
    " align=" << in.align <<
    " endian=" << mem::endian_label()[in.endian];
}

Dbg & operator << (Dbg & dbg, const Data & in) { return dbg << in.info(); }

}
}
}
