/*------------------------------------------------------------------------------
 lib/djv_image/io_dpx.h

 SMPTE Digital Picture Exchange (DPX) Image I/O
 
 References:
 * SMPTE, "SMPTE 268M-2003", http:://www.smpte.org
 * Cinesite, "Conversion of 10-bit Log Film Data To 8-bit Linear or Video Data"

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_IO_DPX_H
#define DJV_IMAGE_IO_DPX_H

#include <djv_image/io.h>

namespace djv {
namespace image {
namespace io {
namespace dpx {

// Information.

static const Str plugin_name = "DPX";
static const List<Str> extension_list = List<Str>() << ".dpx";

enum TAG {
  TAG_SOURCE_OFFSET,
  TAG_SOURCE_CENTER,
  TAG_SOURCE_SIZE,
  TAG_SOURCE_FILE,
  TAG_SOURCE_TIME,
  TAG_SOURCE_INPUT_DEVICE,
  TAG_SOURCE_INPUT_SERIAL,
  TAG_SOURCE_BORDER,
  TAG_SOURCE_PIXEL_ASPECT,
  TAG_SOURCE_SCAN_SIZE,
  TAG_FILM_FORMAT,
  TAG_FILM_FRAME,
  TAG_FILM_SEQUENCE,
  TAG_FILM_HOLD,
  TAG_FILM_FRAME_RATE,
  TAG_FILM_SHUTTER,
  TAG_FILM_FRAME_ID,
  TAG_FILM_SLATE,
  TAG_TV_INTERLACE,
  TAG_TV_FIELD,
  TAG_TV_VIDEO_SIGNAL,
  TAG_TV_SAMPLE_RATE,
  TAG_TV_FRAME_RATE,
  TAG_TV_TIME_OFFSET,
  TAG_TV_GAMMA,
  TAG_TV_BLACK_LEVEL,
  TAG_TV_BLACK_GAIN,
  TAG_TV_BREAK_POINT,
  TAG_TV_WHITE_LEVEL,
  TAG_TV_INTEGRATION_TIMES
};

// Options.

enum TYPE { TYPE_AUTO, U10 };
enum ENDIAN { ENDIAN_AUTO, MSB, LSB };

// Color space.

enum COLOR { COLOR_AUTO, COLOR_RAW, COLOR_PRINT };

class Print {
public:
  Print() : black(95), white(685), gamma(1.7) {}
  int black, white;
  double gamma;
};

class Unprint {
public:
  Unprint() : black(95), white(685), gamma(1.7), soft_clip(0) {}
  int black, white;
  double gamma;
  int soft_clip;
};

// Load.

class Load : public io::Load {
public:

  Load() : _print(false) {}

  Str name() const { return dpx::plugin_name; }
  const List<Str> & extension_list() const { return dpx::extension_list; }
  Str help() const;

  Plugin * copy() const;
  
  struct Options {
    Options() : color(COLOR_AUTO) {}
    COLOR color;
    Unprint print;
  };
  enum OPTION { OPTION_COLOR, OPTION_PRINT };
  bool option(const Str &, List<Str> *);
  List<Str> option(const Str &) const;
  const List<Str> & option_list() const;
  
  void open(const File &, Info *) throw (Error);
  const Image * get(int frame, int proxy) throw (Error);

private:

  void _open(const File &, Info *) throw (Error);
  
  Options _options;
  File _file;
  bool _print;
  std::auto_ptr<Data> _print_lut;
  Cache _cache;
  djv::Io _io;
  Image _image, _image_tmp;
};

// Save.

class Save : public io::Save {
public:
  
  Str name() const { return dpx::plugin_name; }
  const List<Str> & extension_list() const { return dpx::extension_list; }
  Str help() const;
  
  Plugin * copy() const;
  
  struct Options {
    Options() : color(COLOR_PRINT), type(U10), endian(MSB) {}
    COLOR color;
    Print print;
    TYPE type;
    ENDIAN endian;
  };
  enum { OPTION_COLOR, OPTION_PRINT, OPTION_TYPE, OPTION_ENDIAN };
  bool option(const Str &, List<Str> *);
  List<Str> option(const Str &) const;
  const List<Str> & option_list() const;
  
  void open(const File &, const Info &) throw (Error);
  void set(const Image &, int frame) throw (Error);

private:

  void _open(const File &, const tag::Tag &) throw (Error);

  Options _options;
  File _file;
  Info _info;
  djv::Io _io;
  Image _image;
};

// Header functions.

void header_set(djv::Io *, const Str & file, const image::Info &,
  ENDIAN, COLOR
) throw (Error);
void header_update(djv::Io *) throw (Error);
void header_get(djv::Io *, Info *, bool * print = 0) throw (Error);

// Color functions.

void print(pixel::F32_T *, int size, const Print &);
void unprint(pixel::F32_T *, int size, const Unprint &);

Data print_lut(const Print &);
Data unprint_lut(const Unprint &);

// Conversion.

bool serialize(List<Str> *, Print *);
bool serialize(List<Str> *, Unprint *);
List<Str> serialize(const Print &);
List<Str> serialize(const Unprint &);

// Operators.

bool operator == (const Print &, const Print &);
bool operator == (const Unprint &, const Unprint &);

// Labels.

const List<Str> & color_label();
const List<Str> & type_label();
const List<Str> & endian_label();
const List<Str> & tag_label();
const List<Str> & load_option_label();
const List<Str> & save_option_label();

}
}
}
}

#endif
