/*------------------------------------------------------------------------------
 lib/djv_image/io_sgi_load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_sgi_load.cxx"

#include "io_sgi.h"

namespace djv {
namespace image {
namespace io {
namespace _sgi {

Load::Load()
{
  _io.endian(mem::endian() != mem::MSB);
}

void Load::open(const File & in, Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start), info);
  
  if (file::SEQ == _file.type) {
    info->time.start = _file.seq.start;
    info->time.end = _file.seq.end;
  }
}

void Load::_open(const File & in, Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _io.open(in, djv::io::READ);

  uint16_t magic;
  _io.get_u16(&magic);
  DJV_DBG_PRN("magic = " << magic);
  if (magic != 474)
    throw Error(plugin_name, error_unrecognized);

  // Header.
  
  uint8_t storage, bytes;
  uint16_t dimension, width, height, channels;
  uint32_t pixel_min, pixel_max;
  _io.get_u8(&storage);
  _io.get_u8(&bytes);
  _io.get_u16(&dimension);
  _io.get_u16(&width);
  _io.get_u16(&height);
  _io.get_u16(&channels);
  _io.get_u32(&pixel_min);
  _io.get_u32(&pixel_max);
  _io.position(512);

  // Initialize scanline tables.
  
  const int table_size = height * channels;
  DJV_DBG_PRN("rle table size = " << table_size);
  _rle_offset.size(table_size);
  _rle_size.size(table_size);
  if (storage) {
    _io.get_u32(_rle_offset(), table_size);
    _io.get_u32(_rle_size(), table_size);
  } else
    for (uint16_t c = 0; c < channels; ++c)
      for (uint16_t y = 0; y < height; ++y) {
        _rle_offset()[y + height * c] = 512 + (y + height * c) * width * bytes;
        _rle_size()[y + height * c] = width * bytes;
      }
  
  // Information.

  info->size = Vec2i(width, height);

  if (! pixel::pixel(channels, 1 == bytes ? 8 : 16, &info->pixel))
    throw Error(plugin_name, error_unsupported);
  
  info->endian = mem::MSB;
  
  _compress = storage ? true : false;
}

const Image * Load::get(int frame, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start), &info);
  
  // Read.
  
  if (! _compress) {
  
    const uint8_t * p = _io.mmap_p();
    _io.position(data::data_bytes(info), djv::io::OFFSET);
    _image_tmp.set(info, p);
    
  } else {
  
    _image_tmp.set(info);
    const int channels = pixel::channels(info.pixel);
    const int bytes = pixel::channel_bytes(info.pixel);
    for (int c = 0; c < channels; ++c) {
      DJV_DBG_PRN("channel = " << c);
      for (int y = 0; y < info.size.y; ++y) {
        _io.position(_rle_offset()[y + info.size.y * c]);
        rle_unpack(&_io,
          _image_tmp.data() + (c * info.size.y + y) * info.size.x * bytes,
          info.size.x, bytes);
      }
    }
  }
  
  info.size = data::proxy_scale(info.size, proxy);
  info.proxy = proxy;
  _image.set(info);
  data::planar_interleave(_image_tmp, &_image, proxy);
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}
}
}
