/*------------------------------------------------------------------------------
 lib/djv_ui/layout_row.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#undef DJV_DBG
#define _DBG "lib/djv_ui/layout_row.cxx"

#include "layout_row.h"
#include <djv_ui/style.h>

namespace djv {
namespace ui {
namespace layout {

class _Spacer : public Spacer {
public:

  _Spacer(const Vec2i & in) : _size(in) { size_policy(FIXED, FIXED); }

  Vec2i size_hint() const
  {
    return Vec2i(
      (_size.x != -1) ? _size.x : style::style()->spacing(),
      (_size.y != -1) ? _size.y : style::style()->spacing());
  }

private:

  Vec2i _size;
};
  
class _Stretch : public Spacer {};

Row_Base::Row_Base(Item * parent) :
  Layout(parent),
  _stretch(-1)
{}

Row_Base::~Row_Base() {}

void Row_Base::add_spacer(const Vec2i & in, bool stretch)
{
  add(new _Spacer(in));
  if (stretch) this->stretch(int(list().size()) - 1);
}

void Row_Base::add_stretch(bool stretch)
{
  add(new _Stretch);
  if (stretch) this->stretch(int(list().size()) - 1);
}

void Row_Base::stretch(int in)
{
  if (in == _stretch) return;
  _stretch = in;
  //layout();
}

void Row_Base::stretch(const Item * in)
{
  const List<Item *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i)
    if (in == list[i]) {
      stretch(i);
      break;
    }
}

void Row_Base::layout(const List<Item *> & in, List<Box2i> * out,
  const Vec2i & size, DIR dir, int spacing, int stretch
) {
  DJV_DBG_OBJ2(_DBG, "Row_Base::layout");
  DJV_DBG_PRN("size = " << size);
  DJV_DBG_PRN("direction = " << dir);
  DJV_DBG_PRN("stretch = " << stretch);
  
  const uint list_size = in.size();
  out->resize(list_size);
  
  // Calculate minimum sizes.

  int total_length = 0;
  int stretch_count = 0;
  const Item * prev = 0;

  for (uint i = 0; i < list_size; ++i) {
    const Item * item = in[i];
    Vec2i _size;
    
    if (item->visible()) {
    
      _size = item->size_min();
      total_length += _size.e[dir];
      
      if (item->type() != SPACER && prev)
        if (prev->type() != SPACER)
          total_length += spacing;

      if (STRETCH == item->size_policy()[dir] || int(i) == stretch)
        ++stretch_count;

      prev = item;
    }

    (*out)[i] = _size;
    DJV_DBG_PRN(item->dbg_name() << " = " << (*out)[i]);
  }
  DJV_DBG_PRN("total length = " << total_length);
  
  // Expand from minimum sizes to size hints.
  
  for (uint i = 0; (size.e[dir] - total_length) > 0 && i < list_size; ++i) {
    const Item * item = in[i];
    if (item->visible()) {
    
      Vec2i _size = (*out)[i].size;
      const int tmp = _size.e[dir];
      _size.e[dir] = math::min(item->group_hint().e[dir],
        tmp + size.e[dir] - total_length);
      
      (*out)[i] = _size;
      total_length += _size.e[dir] - tmp;
      DJV_DBG_PRN(item->dbg_name() << " = " << (*out)[i]);
    }
  }
  DJV_DBG_PRN("total size = " << total_length);
  
  const int stretch_size = math::max(size.e[dir] - total_length, 0);
  DJV_DBG_PRN("stretch count = " << stretch_count);
  DJV_DBG_PRN("stretch size = " << stretch_size);
  
  // Layout items.

  prev = 0;
  Vec2i position;
  for (uint i = 0; i < list_size; ++i) {
    const Item * item = in[i];
    
    Vec2i _size = (*out)[i].size;
    
    if (item->visible()) {
    
      if (item->type() != SPACER && prev)
        if (prev->type() != SPACER)
          position.e[dir] += spacing;
      
      if (STRETCH == item->size_policy()[dir] || int(i) == stretch) {
        int tmp = 0;
        if (int(i) == stretch)
          tmp = stretch_size;
        else if (-1 == stretch && stretch_count)
          tmp = stretch_size / stretch_count;
        _size.e[dir] += tmp;
      }

      prev = item;
    }
    
    (*out)[i] = Box2i(position, _size);
    position.e[dir] += _size.e[dir];
    DJV_DBG_PRN(item->dbg_name() << " = " << (*out)[i]);
  }
}

Horizontal::Horizontal(Item * parent) :
  Row_Base(parent)
{
  //size_policy(STRETCH, FIXED);
}

void Horizontal::add_spacer(int in, bool stretch)
{
  Row_Base::add_spacer(Vec2i(in, 0), stretch);
}

void Horizontal::dirty()
{
  DJV_DBG_OBJ2(_DBG, "Horizontal::dirty");
  
  Row_Base::dirty();

  layout(list(), &_list, Vec2i(), HORIZONTAL, spacing(), stretch());
  
  int w = 0, h = 0;
  if (_list.size())
    w = _list[_list.size() - 1].x1() + 1 + margin() * 2;
  for (uint i = 0; i < _list.size(); ++i)
    h = math::max(h, _list[i].h);

  _geom_prev = Box2i();
  _size_hint = Vec2i(w, h + margin() * 2);
}

void Horizontal::resize(int x, int y, int w, int h)
{
  if (Box2i(x, y, w, h) == _geom_prev) return;
  _geom_prev = Box2i(x, y, w, h);

  DJV_DBG_OBJ2(_DBG, "Horizontal::resize");
  DJV_DBG_PRN("in = " << Box2i(x, y, w, h));

  Row_Base::resize(x, y, w, h);

  const List<Item *> & list = this->list();
  
  layout(list, &_list, Vec2i(w, h) - margin() * 2, HORIZONTAL, spacing(),
    stretch());
  
  for (uint i = 0; i < _list.size(); ++i) {
    Item * item = list[i];
    DJV_DBG_PRN(list[i]->dbg_name() << " = " << _list[i]);
    
    int _y = _list[i].y, _h = _list[i].h;
    if (STRETCH == item->size_policy()[VERTICAL])
      _h = math::max(h - margin() * 2, _h);
    else
      switch (item->align_policy()[VERTICAL]) {
        case LEFT: break;
        case CENTER: _y += ((h - margin() * 2) - _h) / 2; break;
        case RIGHT: break;
      }
    
    item->resize(x + margin() + _list[i].x, y + margin() + _y, _list[i].w, _h);
  }
}

Vertical::Vertical(Item * parent) :
  Row_Base(parent)
{
  //size_policy(FIXED, STRETCH);
}

void Vertical::add_spacer(int in, bool stretch)
{
  Row_Base::add_spacer(Vec2i(0, in), stretch);
}

void Vertical::dirty()
{
  DJV_DBG_OBJ2(_DBG, "Vertical::dirty");
  
  Row_Base::dirty();

  const List<Item *> & list = this->list();
  layout(list, &_list, Vec2i(), VERTICAL, spacing(), stretch());
  
  int w = 0, h = 0;
  for (uint i = 0; i < _list.size(); ++i)
    w = math::max(w, _list[i].w);
  if (_list.size())
    h = _list[_list.size() - 1].y1() + 1 + margin() * 2;
  
  _geom_prev = Box2i();
  _size_hint = Vec2i(w + margin() * 2, h);
}

void Vertical::resize(int x, int y, int w, int h)
{
  if (Box2i(x, y, w, h) == _geom_prev) return;
  _geom_prev = Box2i(x, y, w, h);
  
  DJV_DBG_OBJ2(_DBG, "Vertical::resize");
  DJV_DBG_PRN("in = " << Box2i(x, y, w, h));

  Row_Base::resize(x, y, w, h);
  
  const List<Item *> & list = this->list();
  layout(list, &_list, Vec2i(w, h) - margin() * 2, VERTICAL, spacing(),
    stretch());
  
  for (uint i = 0; i < _list.size(); ++i) {
    Item * item = list[i];
    DJV_DBG_PRN(item->dbg_name() << " = " << _list[i]);
    
    int _x = _list[i].x, _w = _list[i].w;
    if (STRETCH == item->size_policy()[HORIZONTAL])
      _w = math::max(w - margin() * 2, _w);
    else
      switch (item->align_policy()[HORIZONTAL]) {
        case LEFT: break;
        case CENTER: _x += ((w - margin() * 2) - _w) / 2; break;
        case RIGHT: break;
      }
    
    item->resize(x + margin() + _x, y + margin() + _list[i].y, _w, _list[i].h);
  }
}

}
}
}
