/*------------------------------------------------------------------------------
 plugin/djv_tiff/load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_tiff/load.cxx"

#include "plugin.h"

using namespace image;

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_load()
{
  tiff_init();
  return new Load;
}
}

Load::Load() : _f(0) {}
Load::~Load() { _close(); }

void Load::open(const File & in, image::io::Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start), info);

  if (file::SEQ == _file.type) {
    info->time.start = _file.seq.start;
    info->time.end = _file.seq.end;
  }
}

void Load::_open(const File & in, image::io::Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Open.
  
  _f = TIFFOpen(in.get().c_str(), "r");
  if (! _f)
    throw Error(plugin_name, image::io::error_open);
  
  // Header.

  uint32 width, height;
  uint16 photometric, samples, sample_depth, sample_format;
  uint16 * extra_samples, extra_samples_size;
  uint16 orient, compress, channels;
  
  TIFFGetFieldDefaulted(_f, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetFieldDefaulted(_f, TIFFTAG_IMAGELENGTH, &height);
  TIFFGetFieldDefaulted(_f, TIFFTAG_PHOTOMETRIC, &photometric);
  TIFFGetFieldDefaulted(_f, TIFFTAG_SAMPLESPERPIXEL, &samples);
  TIFFGetFieldDefaulted(_f, TIFFTAG_BITSPERSAMPLE, &sample_depth);
  TIFFGetFieldDefaulted(_f, TIFFTAG_SAMPLEFORMAT, &sample_format);
  TIFFGetFieldDefaulted(_f, TIFFTAG_EXTRASAMPLES, &extra_samples_size,
    &extra_samples);
  TIFFGetFieldDefaulted(_f, TIFFTAG_ORIENTATION, &orient);
  TIFFGetFieldDefaulted(_f, TIFFTAG_COMPRESSION, &compress);
  TIFFGetFieldDefaulted(_f, TIFFTAG_PLANARCONFIG, &channels);
  TIFFGetFieldDefaulted(_f, TIFFTAG_COLORMAP,
    &_colormap[0], &_colormap[1], &_colormap[2]);
  
  // Information.
  
  info->size = Vec2i(width, height);
  
  if (PLANARCONFIG_SEPARATE == channels)
    throw Error(plugin_name, image::io::error_unsupported);

  int pixel = -1;
  switch (photometric) {
    case PHOTOMETRIC_PALETTE: pixel = pixel::RGB_U8; break;
    case PHOTOMETRIC_MINISWHITE:
    case PHOTOMETRIC_MINISBLACK:
    case PHOTOMETRIC_RGB:
      if (32 == sample_depth && sample_format != SAMPLEFORMAT_IEEEFP) break;
      pixel = pixel::pixel(samples, sample_depth);
      break;
  }
  if (-1 == pixel)
    throw Error(plugin_name, image::io::error_unsupported);
  info->pixel = pixel::PIXEL(pixel);

  _compress = compress != COMPRESSION_NONE;
  _palette = PHOTOMETRIC_PALETTE == photometric;
  
  switch (orient) {
    case ORIENTATION_TOPLEFT: info->mirror_y = true; break;
    case ORIENTATION_TOPRIGHT: info->mirror_x = info->mirror_y = true; break;
    case ORIENTATION_BOTRIGHT: info->mirror_x = true; break;
    case ORIENTATION_BOTLEFT: break;
  }
  
  // Image tags.
  
  char * tag;
  if (TIFFGetField(_f, TIFFTAG_ARTIST, &tag))
    info->tag[tag::label()[tag::CREATOR]] = tag;
  if (TIFFGetField(_f, TIFFTAG_IMAGEDESCRIPTION, &tag))
    info->tag[tag::label()[tag::DESCRIPTION]] = tag;
  if (TIFFGetField(_f, TIFFTAG_COPYRIGHT, &tag))
    info->tag[tag::label()[tag::COPYRIGHT]] = tag;
  if (TIFFGetField(_f, TIFFTAG_DATETIME, &tag))
    info->tag[tag::label()[tag::TIME]] = tag;
}

void Load::_close()
{
  if (_f) { TIFFClose(_f); _f = 0; }
}

const Image * Load::get(int frame, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  image::io::Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start), &info);
  
  // Read.
  
  Image * data = proxy ? &_image_tmp : &_image;
  data->set(info);
  
  for (int y = 0; y < info.size.y; ++y) {
    if (TIFFReadScanline(_f, (tdata_t *)data->data(0, y), y) == -1)
      throw Error(plugin_name, image::io::error_read);
    if (_palette)
      palette_unpack(data->data(0, y), info.size.x,
        pixel::channel_bytes(info.pixel),
        _colormap[0], _colormap[1], _colormap[2]);
  }
  _close();

  if (proxy) {
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_image_tmp, &_image, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}
